// @dart = 2.8

import 'dart:async';

import 'package:yaml/yaml.dart';

import '../base/common.dart';
import '../globals.dart';
import '../runner/faircli_command.dart';
import '../util/util.dart';

const fairVersion = '^3.0.0';
const fairCompiler = '^1.5.0';
const fairVersionVersion = '^3.0.0';
const buildRunnerVersion = '^2.0.0';
const fairPushyVersion = '''
    git:
      url: https://github.com/wuba/FairPushy.git
      ref: main
      path: sdk''';

class ImportCommand extends FaircliCommand {
  ImportCommand() {
    argParser.addOption(
      'description',
      defaultsTo: 'Import Fair dependency to your carrier project.',
      help: 'The description to use for your Fair project.',
    );

    argParser.addOption(
      'directory',
      abbr: 'd',
      defaultsTo: '',
      help: 'flutter project directory',
    );
  }

  @override
  final String name = 'import';

  @override
  final String description = 'Import Fair dependency to your carrier project.';

  @override
  Future<FaircliCommandResult> runCommand() async {
    String directory = argResults['directory'];
    if (directory.isEmpty) {
      throwToolExit(
          'Please use "faircli import -d xxx" to import Fair dependency to your carrier project');
    }
    var projectDir = fs.directory(directory);
    if (!projectDir.existsSync()) {
      throwToolExit('The input directory is not exists');
    }
    cliPrint('Start to import Fair dependency, wait a moment please...');
    // add Fair dependency
    var pubspecFile = fs.file(fs.path.join(directory, 'pubspec.yaml'));
    if (!pubspecFile.existsSync()) {
      throwToolExit(
          'The directory is not a Flutter project, "pubspec.yaml" is not exists');
    }
    YamlMap yamlData;
    try {
      yamlData = loadYaml(pubspecFile.readAsStringSync());
    } catch (e) {
      throwToolExit('loadYaml failed: $e');
    }
    var dependenciesEndLine = _getNodeYamlEndLine(yamlData, 'dependencies');
    var devDependenciesEndLine =
        _getNodeYamlEndLine(yamlData, 'dev_dependencies');
    var dependencyOverridesEndLine =
        _getNodeYamlEndLine(yamlData, 'dependency_overrides');
    var pubspecStr = StringBuffer();
    var lineIndex = 1;
    var ignoreLines = <List<int>>[];
    ignoreLines
        .add(_yamlExistsLinesDependency(yamlData, 'dependencies', 'fair'));
    ignoreLines
        .add(_yamlExistsLinesDependency(yamlData, 'dependencies', 'fair_pushy'));
    ignoreLines.add(_yamlExistsLinesDependency(
        yamlData, 'dev_dependencies', 'fair_compiler'));
    ignoreLines.add(_yamlExistsLinesDependency(
        yamlData, 'dev_dependencies', 'build_runner'));
    ignoreLines.add(_yamlExistsLinesDependency(
        yamlData, 'dependency_overrides', 'fair_version'));
    pubspecFile.readAsLinesSync().forEach((line) {
      if (dependenciesEndLine != -1 && lineIndex == dependenciesEndLine) {
        if (!_isIgnoreLines(ignoreLines, lineIndex)) {
          pubspecStr.writeln(line);
        }
        _writeFairDependency(yamlData, pubspecStr);
      } else if (devDependenciesEndLine != -1 &&
          lineIndex == devDependenciesEndLine) {
        if (!_isIgnoreLines(ignoreLines, lineIndex)) {
          pubspecStr.writeln(line);
        }
        _writeFairDevDependency(yamlData, pubspecStr);
      } else if (dependencyOverridesEndLine != -1 &&
          lineIndex == dependencyOverridesEndLine) {
        if (!_isIgnoreLines(ignoreLines, lineIndex)) {
          pubspecStr.writeln(line);
        }
        _writeFairDevDependencyOverrides(yamlData, pubspecStr);
      } else if (!_isIgnoreLines(ignoreLines, lineIndex)) {
        pubspecStr.writeln(line);
      }
      lineIndex++;
    });
    if (dependenciesEndLine == -1) {
      pubspecStr.writeln('dependencies:');
      _writeFairDependency(yamlData, pubspecStr);
    }
    if (devDependenciesEndLine == -1) {
      pubspecStr.writeln('dev_dependencies:');
      _writeFairDevDependency(yamlData, pubspecStr);
    }
    if (dependencyOverridesEndLine == -1) {
      pubspecStr.writeln('dependency_overrides:');
      _writeFairDevDependencyOverrides(yamlData, pubspecStr);
    }
    pubspecFile.writeAsStringSync(pubspecStr.toString());
    cliPrint('Import Fair dependency success');
    cliPrint('''
Then you should modify the 'main.dart' like this:

void main() {
  WidgetsFlutterBinding.ensureInitialized();

  FairApp.runApplication(
    _getApp(),
    plugins: {
    },
  );
}

dynamic _getApp() => FairApp(
  modules: {
  },
  delegate: {
  },
  child: MaterialApp(
    home: FairWidget(
            name: 'DynamicWidget',
            path: 'assets/bundle/lib_src_page_dynamic_widget.fair.json',
            data: {"fairProps": json.encode({})}),
  ),
);

Enjoy~
    ''');
    return null;
  }

  void _writeFairDependency(YamlMap yamlData, StringBuffer pubspecStr) {
    pubspecStr.writeln('  fair: $fairVersion');
    pubspecStr.writeln('  fair_pushy: \n$fairPushyVersion');
  }

  void _writeFairDevDependencyOverrides(
      YamlMap yamlData, StringBuffer pubspecStr) {
    pubspecStr.writeln('  fair_version: $fairVersionVersion');
  }

  void _writeFairDevDependency(YamlMap yamlData, StringBuffer pubspecStr) {
    pubspecStr.writeln('  fair_compiler: $fairCompiler');
    pubspecStr.writeln('  build_runner: $buildRunnerVersion');
  }

  _yamlExistsLinesDependency(YamlMap yamlData, String node, String package) {
    if (!yamlData.containsKey(node) || yamlData[node][package] == null) {
      return [-1, -1];
    }
    var nodeMap = yamlData.nodes[node];
    if (nodeMap is YamlMap) {
      var packageNode = nodeMap.nodes[package];
      if (packageNode is YamlMap) {
        return [packageNode.span.start.line, packageNode.span.end.line];
      } else {
        return [packageNode.span.start.line + 1, packageNode.span.end.line + 1];
      }
    }
    return [-1, -1];
  }

  _getNodeYamlEndLine(YamlMap yamlData, String node) {
    if (!yamlData.containsKey(node)) {
      return -1;
    }
    if (yamlData.nodes[node] is YamlMap) {
      var nodeMap = (yamlData.nodes[node] as YamlMap);
      var last = nodeMap.nodes.values.last;
      if (last != null) {
        if (last is YamlMap) {
          return last.span.end.line;
        } else {
          return last.span.end.line + 1;
        }
      }
      return yamlData.nodes[node].span.end.line;
    }
    return yamlData.nodes[node].span.end.line + 1;
  }

  _isIgnoreLines(List<List<int>> ignoreLines, int lineIndex) {
    for (var ignoreLine in ignoreLines) {
      if (lineIndex >= ignoreLine[0] && lineIndex <= ignoreLine[1]) {
        return true;
      }
    }
    return false;
  }
}
